/*
 * NSView subclass for Mac OS X rootless X server
 */
/* $XFree86: xc/programs/Xserver/hw/darwin/bundle/XView.m,v 1.5 2001/11/09 00:12:38 torrey Exp $ */

#include <ApplicationServices/ApplicationServices.h>

#import "XView.h"
#include "fakeBoxRec.h"
#include "unistd.h"
#include "quartzCommon.h"

static const void *infobytes(void *info)
{
    return info;
}


static unsigned long *shapeBits = NULL, *tempBits = NULL;
static int shapeWidth = 0;
static int shapeHeight = 0;
static int shapeDrawing = 0;

static void reallocShapeBits(NSSize minSize)
{
    unsigned long *oldBits;
    int oldWidth, oldHeight;
    
    if (shapeWidth*shapeHeight < minSize.width*minSize.height) {
        oldBits=shapeBits;
        oldWidth=shapeWidth;
        oldHeight=shapeHeight;
        shapeWidth = minSize.width;
        shapeHeight = minSize.height;
        if (shapeBits) free(shapeBits);
        shapeBits = (unsigned long *) malloc(4 * shapeWidth * shapeHeight);
//        NSLog(@"Reallocated shape bits... %#lx %d x %d (%#lx %d x %d)",shapeBits,shapeWidth,shapeHeight,oldBits,oldWidth,oldHeight);
    }
}


@implementation XView

- (id)initWithFrame:(NSRect)aRect
{
    self = [super initWithFrame:aRect];
    if (!self) return nil;

//    mShaped = NO;
    mShaped = YES;
    mBitsPerSample = 8;
    mSamplesPerPixel = 3;
    mDepth = mBitsPerSample * mSamplesPerPixel;
    mBitsPerPixel = 32;
    mRows = 0;
    mBits = nil;
    [self allocBitsForSize:aRect.size isRoot:YES];
    mDimmed = NO;
    mInhibitShadowChange = NO;
    mMiniature = NO;
    mDrawAlpha = 0xff000000;
    mTranslucentBG = 0xffffffff;
    mWaitForDrawRect = NO;
    mInhibitUpdate = NO;
// this will be filled in with an ID from OroborOSX, so XDarwin can signal a specific window
    mHackID = 0;
    myFrame=aRect;
    //[self setAutoresizingMask:(NSViewMinYMargin | NSViewMaxXMargin)];
    //[self setAutoresizingMask:(NSViewMinYMargin | NSViewMaxXMargin)];
    
    return self;
}

- (void)dealloc
{
    if (mBits) free(mBits);
    [super dealloc];
}

- (void)copyAllToShapeBits
{
    fakeBoxRec bounds = {[self frame].origin.x, [self frame].origin.y, [self frame].size.width,
                         [self frame].size.height};
    [self copyToShapeBits:&bounds count:1];
}

- (void)drawRect:(NSRect)aRect
{
    fakeBoxRec bounds = {aRect.origin.x, aRect.origin.y, aRect.size.width,
                         aRect.size.height};
    BOOL	wasMin;
    
    // Never draw here.
//    NSLog(@"HACKID %d : Wants to draw... %#lx %d %d",mHackID,mBits,mWaitForDrawRect,mMiniature);
    
    if(!mMiniature && !mWaitForDrawRect) return;
    
    wasMin = mMiniature;
    
        while(shapeDrawing) {
            NSLog(@"HACKID %d : sleeping in drawRect until shapeBits freed... %#lx",mHackID,mBits);   
            usleep(10000);
        }
        shapeDrawing = 1;
//        reallocShapeBits(aRect.size);
        mMiniature=NO; // temporarily allow copyToScreen to draw to window
        // copy everything from X11 to temp
        // copy everything from temp to screen
        [self copyToShapeBits:&bounds count:1];
        if(!mWaitForDrawRect) {
//            NSLog(@"HACKID %d : clearing in drawRect... %#lx",mHackID,mBits);   
            [self lockFocus];
            [[NSColor clearColor] set];
            NSRectFill(aRect);
            [self unlockFocus];
        }
        [self lockFocus];
        [self copyToScreen:&bounds count:1 fromTemp:YES];
//        NSLog(@"HACKID %d : done copy in drawRect... %#lx",mHackID,mBits);   
        [[NSGraphicsContext currentContext] flushGraphics];
        [self unlockFocus];
        mMiniature=wasMin;
        shapeDrawing = 0;
//        [[self window] resetAlphaValue];
        mWaitForDrawRect=NO;
}

- (void)setWaitForDrawRect:(BOOL)val
{
    mWaitForDrawRect = val;
}

- (BOOL)getWaitForDrawRect
{
    return mWaitForDrawRect;
}

- (BOOL)isFlipped
{
    return NO;
}

- (BOOL)isOpaque
{
//    return YES;
    return !mShaped;
}

- (BOOL)acceptsFirstResponder
{
    return !mMiniature;
}

- (BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
    return !mMiniature;
}

- (BOOL)shouldDelayWindowOrderingForEvent:(NSEvent *)theEvent
{
    return !mMiniature;
}


- (void)mouseEntered:(NSEvent *)anEvent
{
    xwinHackID[hackIDPos] = mHackID;
    //NSLog(@"HACKID %d : <<<<<<<< Sending entered notification\n",mHackID);
    QuartzMessageMainThread(kQuartzWinEntered+hackIDPos);
    hackIDPos++;
    if(hackIDPos>=kDockPosMax) hackIDPos=0;
}

- (void)mouseExited:(NSEvent *)anEvent
{
    xwinHackID[hackIDPos] = mHackID;
    //NSLog(@"HACKID %d : <<<<<<<< Sending exited notification\n",mHackID);
    QuartzMessageMainThread(kQuartzWinExited+hackIDPos);
    hackIDPos++;
    if(hackIDPos>=kDockPosMax) hackIDPos=0;
}

- (void)mouseDown:(NSEvent *)anEvent
{
//    if(mMiniature) return;
    // Only X is allowed to restack windows.
//    NSLog(@"HACKID %d : Got mouse down... (%d)",mHackID,can_come_in_front);
    [NSApp preventWindowOrdering];
    if (! [NSApp isActive] && can_come_in_front==1)	// used when OroborOSX is in menu select
        [NSApp activateIgnoringOtherApps:YES];
//    [[self window] setAcceptsMouseMovedEvents:YES];
    [[self nextResponder] mouseDown:anEvent];
}

- (void)mouseUp:(NSEvent *)anEvent
{
//    NSLog(@"HACKID %d : Got mouse up... (%d)",mHackID,can_come_in_front);
//    if(mMiniature) return;
    // Bring app to front if necessary
    // Don't bring app to front in mouseDown; mousedown-mouseup is too
    // long and X gets what looks like a mouse drag.
    if (! [NSApp isActive] ) {
        if(can_come_in_front==2)	// not currently used
            [NSApp activateIgnoringOtherApps:YES];
        else if(can_come_in_front==3) {	// original XDarwin behaviour
            [NSApp activateIgnoringOtherApps:YES];
            [NSApp arrangeInFront:nil];
        }
    }

    [[self nextResponder] mouseDown:anEvent];
//    [[self nextResponder] mouseUp:anEvent];
}

//- (NSRect)frame
//{
//    return myFrame;
//}

//- (void)setFrame:(NSRect)aRect
//{
//    fakeBoxRec bounds = {0, 0, aRect.size.width, aRect.size.height};

//    if(mHackID) {
//        NSLog(@"HACKID %d : removing tracking rect in setFrame... %d",mHackID,mRectTag);   
//        [self removeTrackingRect:mRectTag];
//    }
//    myFrame=aRect;
//    [super setFrame:aRect];
//    if(mHackID) {
//        mRectTag = [self addTrackingRect:[self frame] owner:self userData:nil assumeInside:NO];
//        NSLog(@"HACKID %d : changed tracking rect in setFrame... %d",mHackID,mRectTag);
//    }
//}

// setFrame goes through here too.
//- (void)setFrameSize:(NSSize)newSize
//{
//    fakeBoxRec bounds = {0, 0, newSize.width, newSize.height};
    
/*    if(mHackID) {
        NSLog(@"HACKID %d : removing tracking rect in setFrameSize... %d",mHackID,mRectTag);   
        [self removeTrackingRect:mRectTag];
    }*/
//    [self allocBitsForSize:newSize];
//        [self lockFocus];
//        [[NSColor clearColor] set];
//        NSRectFill([self frame]);
//        [self unlockFocus];
//    [super setFrameSize:newSize];
    //myFrame.size=newSize;
/*    if(mHackID) {
        mRectTag = [self addTrackingRect:[self frame] owner:self userData:nil assumeInside:NO];
        NSLog(@"HACKID %d : changed tracking rect in setFrameSize... %d",mHackID,mRectTag);
    }*/
    
        // copy everything from X11 to temp
        // copy everything from temp to screen
/*        while(shapeDrawing) {
            NSLog(@"HACKID %d : sleeping in setFrameSize until shapeBits freed... %#lx",mHackID,mBits);   
            usleep(10000);
        }
        shapeDrawing = 1;
//        [self lockFocus];
        [self copyToShapeBits:&bounds count:1];
        [self copyToScreen:&bounds count:1 fromTemp:YES];
        [[NSGraphicsContext currentContext] flushGraphics];
//        [self unlockFocus];
        shapeDrawing = 0;
*/
//}

// Reallocate bits.
- (void)allocBitsForSize:(NSSize)newSize isRoot:(BOOL)isRoot
{
    unsigned long *pix;
    int i;
    
    while(mWaitForDrawRect) {
        NSLog(@"HACKID %d : sleeping in allocBitsForSize until drawrect done... %#lx",mHackID,mBits);   
        usleep(10000);
    }
    while(shapeDrawing) {
        NSLog(@"HACKID %d : sleeping in allocBitsForSize until shapeBits freed... %#lx",mHackID,mBits);   
        usleep(10000);
    }
    shapeDrawing = 1;
    
    mBytesPerRow = newSize.width * mBitsPerPixel / 8;
    mRows = newSize.height;

        pix=mBits;
        if(mBits) {
            free(mBits);
            mBits = malloc(mBytesPerRow * newSize.height);
//            NSLog(@"HACKID %d : reallocated mBits: %#lx (%#lx)",mHackID,mBits,pix);
        } else {
            mBits = malloc(mBytesPerRow * newSize.height);
//            NSLog(@"HACKID %d : newly allocated mBits: %#lx",mHackID,mBits);
            pix = mBits;
            for(i=0;i<newSize.width*newSize.height;i++)
                *pix++ = 0;
        }
        if(!isRoot) reallocShapeBits(newSize);
        shapeDrawing = 0;
}

- (char *)bits
{
    return mBits;
}

- (void)getBits:(char **)bits
       rowBytes:(int *)rowBytes
          depth:(int *)depth
   bitsPerPixel:(int *)bpp
{
    *bits = mBits;
    *rowBytes = mBytesPerRow;
    *depth = mDepth;
    *bpp = mBitsPerPixel;
}

- (void)refreshRects:(fakeBoxRec *)rectList count:(int)count
{
    fakeBoxRec *r;
    fakeBoxRec *end;
    NSRect aRect;

//    if(mMiniature) return;
        
    if (!mShaped) {
        [self lockFocus];
        [self copyToScreen:rectList count:count fromTemp:NO];
        [[NSGraphicsContext currentContext] flushGraphics];
        [self unlockFocus];
    } else {
        while(shapeDrawing) {
            NSLog(@"HACKID %d : Sleeping in refreshRects while shape bits in use... %#lx",mHackID,mBits);
            usleep(10000);
        }
        shapeDrawing = 1;
        // only need to clear if the drawing might not be completely opaque
        if(mDrawAlpha!=0xff000000 || (mTranslucentBG & 0xff000000)!=0xff000000) {
//            NSLog(@"HACKID %d : Clearing rects in refreshRects... %d",mHackID,count);
            [self lockFocus];
            for (r = rectList, end = rectList + count; r < end; r++) {
                NSRect nsr = {{r->x1, r->y1}, {r->x2-r->x1, r->y2-r->y1}};
                [[NSColor clearColor] set];
                nsr.origin.y = [self frame].size.height - nsr.origin.y - nsr.size.height;
//                mWaitForDrawRect=YES;
//                [[self window] setAutodisplay:YES];
//                [self setNeedsDisplayInRect:nsr];
//                while(mWaitForDrawRect) {
//                    NSLog(@"Waiting for drawRect in refreshRects");
//                    usleep(100);
//                }
//                [[self window] setAutodisplay:NO];
                NSRectFill(nsr);
            }
//            return;
            [self unlockFocus];
        }
        [self lockFocus];
//        for (r = rectList, end = rectList + count; r < end; r++) {
//            [self copyToShapeBits:r count:1];
//            [self copyToScreen:r count:1 fromTemp:YES];
//        }
        [self copyToScreenViaShape:rectList count:count];
        [[NSGraphicsContext currentContext] flushGraphics];
        [self unlockFocus];
        shapeDrawing = 0;
    }
}

// eraseRects are OUTSIDE the new shape
- (void)reshapeRects:(fakeBoxRec *)eraseRects count:(int)count
{
    fakeBoxRec bounds = {0, 0, [self frame].size.width,
                         [self frame].size.height};
    fakeBoxRec *r;
    fakeBoxRec *end;
    NSRect aRect;

//    if(mMiniature) return;
    
    if (count == 0  &&  !mShaped) {
        [self refreshRects:&bounds count:1];
    } else {
        // View is shaped, or used to be shaped.
        // Shaped windows never become unshaped.
        // (Mac OS X 10.0.4 does not allow transparent windows
        // to become opaque.)

        mShaped = YES;

        while(shapeDrawing) {
            NSLog(@"HACKID %d : Sleeping in reshapeRects while shape bits in use... %#lx",mHackID,mBits);
            usleep(10000);
        }
        shapeDrawing = 1;
        reallocShapeBits([self frame].size);
        // Magic. 10.0.4 and 10.1 both require the alpha channel to be
        // cleared explicitly. 10.0.4 additionally requires the view to
        // be unlocked between this and the drawing code below.
            [self lockFocus];
            for (r = eraseRects, end = eraseRects + count; r < end; r++) {
                NSRect nsr = {{r->x1, r->y1}, {r->x2-r->x1, r->y2-r->y1}};
                [[NSColor clearColor] set];
                nsr.origin.y = [self frame].size.height - nsr.origin.y - nsr.size.height;
                NSRectFill(nsr);
            }
            [self unlockFocus];
        
        // copy everything from X11 to temp
        // erase eraseRects from temp
        // copy everything from temp to screen
        [self eraseFromShapeBits:eraseRects count:count];
//        [self setNeedsDisplay:YES];
        if(!mInhibitUpdate) {
//            NSLog(@"Updating whole window for shape");
            [self copyToShapeBits:&bounds count:1];
            [self lockFocus];
            [self copyToScreen:&bounds count:1 fromTemp:YES];
            [[NSGraphicsContext currentContext] flushGraphics];
            [self unlockFocus];
        }
//        } else
//            NSLog(@"Not updating whole window for shape");
        shapeDrawing = 0;
    }
}


- (void)eraseFromShapeBits:(fakeBoxRec *)rectList count:(int)count
{
    unsigned long *dst = NULL, *bdst = NULL; // don't assign yet
    int dstWidth = 0; // don't assign yet
    fakeBoxRec *r;
    fakeBoxRec *end;

//    if(mMiniature) return;

    assert(mBitsPerPixel == 32);

//    reallocShapeBits([self frame].size);
    dst = shapeBits;
    bdst = (unsigned long *)mBits;
//    dstWidth = shapeWidth;
    dstWidth = mBytesPerRow / 4;

    for (r = rectList, end = rectList + count; r < end; r++) {
        unsigned long *dstline = dst + dstWidth*r->y1 + r->x1;
        unsigned long *bdstline = bdst + dstWidth*r->y1 + r->x1;
        int h = r->y2 - r->y1;

        while (h--) {
            unsigned long *dstp  = dstline;
            unsigned long *bdstp = bdstline;
            int w = r->x2 - r->x1;

            while (w--) {
  //              *dstp++  = 0x00000000;
                *bdstp++ = 0xff000000;
            }
            dstline  += dstWidth;
            bdstline += dstWidth;
        }
    }
}

// assumes X11 bits and shape bits are 32-bit
- (void)copyToShapeBits:(fakeBoxRec *)rectList count:(int)count
{
    unsigned long *src = (unsigned long *) mBits;
    unsigned long *dst = NULL; // don't assign yet
    unsigned long newval;
    int srcWidth = mBytesPerRow / 4;
    int dstWidth = 0; // don't assign yet
    fakeBoxRec *r;
    fakeBoxRec *end;

//    if(mMiniature) return;
    if(count<=0) return;
    
    assert(mBitsPerPixel == 32);

//    reallocShapeBits([self frame].size);
    dst = shapeBits;
//    dstWidth = shapeWidth;
    if(count==1) {
        dstWidth = rectList->x2 - rectList->x1;
        if(dstWidth>srcWidth)
            dstWidth=srcWidth;
    } else
        dstWidth = srcWidth;
    
    for (r = rectList, end = rectList + count; r < end; r++) {
        unsigned long *srcline = src + srcWidth*r->y1 + r->x1;
        unsigned long *dstline = dst + dstWidth*r->y1 + r->x1;
        int h = r->y2 - r->y1;
        if(h>mRows)
            h=mRows;
        if(count==1) dstline = dst;

        while (h--) {
            unsigned long *srcp = srcline;
            unsigned long *dstp = dstline;
            int w = r->x2 - r->x1;
            if(w>dstWidth)
                w=dstWidth;
                
            if(mDimmed) {
                while(w--) {
                    newval = *srcp++;
                    if(newval==0xff000000)
                        *dstp++ = 0;
                    else {
                    newval = newval & 0x00ffffff;
                    if(newval==(mTranslucentBG & 0x00ffffff)) {
                        *dstp++ = (((newval & 0x00fefefe) + 0x00808080) >> 1) | (mTranslucentBG & 0xff000000);
                    } else
                        *dstp++ = (((newval & 0x00fefefe) + 0x00808080) >> 1) | mDrawAlpha;
                    }
                }
//                    newval = *srcp++ & 0x00fefefe;
//                    newval = (newval << 2) + (newval & 0x00f8f8f8);
//                    *dstp++ = ((newval + 0x01818180) >> 3) | 0xff000000;
            } else {
                while(w--) {
                    newval = *srcp++;
                    if(newval==0xff000000)
                        *dstp++ = 0;
                    else {
                    newval = newval & 0x00ffffff;
                    if(newval==(mTranslucentBG | 0x00ffffff)) {
                        *dstp++ = mTranslucentBG;
                    } else
                        *dstp++ = newval | mDrawAlpha;
                    }
                }
            }
//            while (w--) {
//                *dstp++ = (((*srcp++ & 0x00fefefe)+0x00808080) >> 1) | 0xff000000;
//                newval = (*srcp & 0x00ffffff) | 0xc0000000;
//                if(*dstp != newval)
//                    *dstp++ = (*srcp++ & 0x00ffffff) | 0xc0000000;
//                else {
//                    *dstp++ = 0x00000000;
//                    *srcp++;
//                }
//            }
            srcline += srcWidth;
            dstline += dstWidth;
        }
    }
}


// Copy boxes to the screen from the per-window pixmaps where X draws.
// rectList is in local, X-flipped coordinates.
- (void)copyToScreen:(fakeBoxRec *)rectList count:(int)count
            fromTemp:(BOOL)copyFromTemp
{
    unsigned char *offsetbits;
    fakeBoxRec *r;
    fakeBoxRec *end;
    NSRect bounds;
    char *srcBits;
    int bytesPerRow;
    int bitsPerPixel;
    int bitsPerSample;
    CGImageAlphaInfo alpha;

    CGContextRef destCtx;
    CGColorSpaceRef colorSpace;
    const CGDataProviderDirectAccessCallbacks cb = { infobytes, NULL, NULL, NULL };

    //if( mMiniature ) return;
//    if( mMiniature )
//        NSLog(@"Drawing for miniaturised window (%d)...",mHackID);
        
    destCtx = (CGContextRef)[[NSGraphicsContext currentContext]
                                                graphicsPort];
                                                
//    destCtx = [NSGraphicsContext currentContext];
    colorSpace = CGColorSpaceCreateDeviceRGB();
    // fixme colorspace leaks?
        
    if (copyFromTemp) {
        // shapeBits assumed to be 32-bit
        srcBits = (char *)shapeBits;
//        bytesPerRow = 4 * shapeWidth;
        if(count==1) {
            bytesPerRow = (rectList->x2 - rectList->x1) * 4;
            if(bytesPerRow>mBytesPerRow)
                bytesPerRow = mBytesPerRow;
        } else
            bytesPerRow = mBytesPerRow;
            
        bitsPerPixel = 32;
        bitsPerSample = 8;
        alpha = kCGImageAlphaPremultipliedFirst; // premultiplied ARGB (not meant to use -see below)
//        alpha = kCGImageAlphaFirst; // non-premultiplied (should use this but crashes OSX10.0???)
    } else {
        srcBits = mBits;
        bytesPerRow = mBytesPerRow;
        bitsPerPixel = mBitsPerPixel;
        bitsPerSample = mBitsPerSample;
        alpha = kCGImageAlphaNoneSkipFirst; // xRGB
    }

//    bounds = [self frame];
    bounds.origin.x = bounds.origin.y = 0;
    bounds.size.width = mBytesPerRow/4;
    bounds.size.height = mRows;

    for (r = rectList, end = rectList + count; r < end; r++) {
        NSRect nsr = {{r->x1, r->y1}, {r->x2-r->x1, r->y2-r->y1}};
        CGRect destRect;
        CGDataProviderRef dataProviderRef;
        CGImageRef imageRef;

        // Clip to window
        // (bounds origin is (0,0) so it can be used in either flip)
        // fixme is this necessary with pixmap-per-window?
        nsr = NSIntersectionRect(nsr, bounds);

        // Disallow empty rects
        if (nsr.size.width <= 0  ||  nsr.size.height <= 0) continue;

        if(count==1)
            offsetbits = srcBits;
        else
            offsetbits = srcBits + (int)(nsr.origin.y * bytesPerRow +
                                        nsr.origin.x * bitsPerPixel/8);

        // Flip r to Cocoa-flipped
        nsr.origin.y = [self frame].size.height - nsr.origin.y - nsr.size.height;
        destRect = CGRectMake(nsr.origin.x, nsr.origin.y,
                              nsr.size.width, nsr.size.height);

        dataProviderRef = CGDataProviderCreateDirectAccess(offsetbits,
                                destRect.size.height * bytesPerRow, &cb);
//        dataProviderRef = CGDataProviderCreateWithData(NULL,offsetbits,
//                                destRect.size.height * bytesPerRow,NULL);

        imageRef = CGImageCreate(destRect.size.width, destRect.size.height,
                                 bitsPerSample, bitsPerPixel, bytesPerRow,
                                 colorSpace, alpha, dataProviderRef, NULL,
//                                 0, kCGRenderingIntentDefault);
                                 1, kCGRenderingIntentDefault);

        CGContextDrawImage(destCtx, destRect, imageRef);
        CGImageRelease(imageRef);
        CGDataProviderRelease(dataProviderRef);
    }
}

// Combined update routine, in an attempt to make it slightly faster...
// assumes X11 bits and shape bits are 32-bit
- (void)copyToScreenViaShape:(fakeBoxRec *)rectList count:(int)count
{
    unsigned long *src = (unsigned long *) mBits;
    unsigned long *dst = NULL; // don't assign yet
    unsigned long newval;
    int srcWidth = mBytesPerRow / 4;
    int dstWidth = 0; // don't assign yet
    fakeBoxRec *r;
    fakeBoxRec *end;
    unsigned char *offsetbits;
    NSRect bounds;
    char *srcBits;
    int bytesPerRow;
    int bitsPerPixel;
    int bitsPerSample;
    CGImageAlphaInfo alpha;

    CGContextRef destCtx;
    CGColorSpaceRef colorSpace;
    const CGDataProviderDirectAccessCallbacks cb = { infobytes, NULL, NULL, NULL };

    //if( mMiniature ) return;
//    if( mMiniature )
//        NSLog(@"Drawing for miniaturised window (%d)...",mHackID);
        
    destCtx = (CGContextRef)[[NSGraphicsContext currentContext] graphicsPort];
                                                
//    destCtx = [NSGraphicsContext currentContext];
    colorSpace = CGColorSpaceCreateDeviceRGB();
    // fixme colorspace leaks?
        
    // shapeBits assumed to be 32-bit
    srcBits = (char *)shapeBits;
    bytesPerRow = mBytesPerRow;
    bitsPerPixel = 32;
    bitsPerSample = 8;
    alpha = kCGImageAlphaPremultipliedFirst; // premultiplied ARGB (not meant to use -see below)
//    alpha = kCGImageAlphaFirst; // non-premultiplied (should use this but crashes OSX10.0???)

    bounds = [self frame];
    bounds.origin.x = bounds.origin.y = 0;

//    if(mMiniature) return;
    if(count<=0) return;
    
    assert(mBitsPerPixel == 32);

    //reallocShapeBits([self frame].size);
    dst = shapeBits;
    
    for (r = rectList, end = rectList + count; r < end; r++) {
    
        NSRect nsr = {{r->x1, r->y1}, {r->x2 - r->x1, r->y2-r->y1}};
        CGRect destRect;
        CGDataProviderRef dataProviderRef;
        CGImageRef imageRef;
        unsigned long *srcline = src + srcWidth*r->y1 + r->x1;
        unsigned long *dstline = dst;
        int h = r->y2 - r->y1;
        dstWidth = r->x2 - r->x1;
//        dstWidth = shapeWidth;

        // Clip to window
        // (bounds origin is (0,0) so it can be used in either flip)
        // fixme is this necessary with pixmap-per-window?
        nsr = NSIntersectionRect(nsr, bounds);

        // Disallow empty rects
        if (nsr.size.width <= 0  ||  nsr.size.height <= 0) continue;

        while (h--) {
            unsigned long *srcp = srcline;
            unsigned long *dstp = dstline;
            int w = dstWidth;

            if(mDimmed) {
                while(w--) {
                    newval = *srcp++;
                    if(newval==0xff000000)
                        *dstp++ = 0;
                    else {
                        newval = newval & 0x00ffffff;
                        if(newval==(mTranslucentBG & 0x00ffffff))
                            *dstp++ = (((newval & 0x00fefefe) + 0x00808080) >> 1)
                                    | (mTranslucentBG & 0xff000000);
                        else
                            *dstp++ = (((newval & 0x00fefefe) + 0x00808080) >> 1)
                                    | mDrawAlpha;
                    }
                }
            } else {
                while(w--) {
                    newval = *srcp++;
                    if(newval==0xff000000)
                        *dstp++ = 0;
                    else {
                        newval = newval & 0x00ffffff;
                        if(newval==(mTranslucentBG | 0x00ffffff))
                            *dstp++ = mTranslucentBG;
                        else
                            *dstp++ = newval | mDrawAlpha;
                    }
                }
            }
            srcline += srcWidth;
            dstline += dstWidth;
        }

        offsetbits = srcBits;
        bytesPerRow = dstWidth * 4;

        // Flip r to Cocoa-flipped
        nsr.origin.y = bounds.size.height - nsr.origin.y - nsr.size.height;
        destRect = CGRectMake(nsr.origin.x, nsr.origin.y,
                              nsr.size.width, nsr.size.height);

        dataProviderRef = CGDataProviderCreateDirectAccess(offsetbits,
                                destRect.size.height * bytesPerRow, &cb);
//        dataProviderRef = CGDataProviderCreateWithData(NULL,offsetbits,
//                                destRect.size.height * bytesPerRow,NULL);

        imageRef = CGImageCreate(destRect.size.width, destRect.size.height,
                                 bitsPerSample, bitsPerPixel, bytesPerRow,
                                 colorSpace, alpha, dataProviderRef, NULL,
//                                 0, kCGRenderingIntentDefault
                                 1, kCGRenderingIntentDefault);

        CGContextDrawImage(destCtx, destRect, imageRef);
        CGImageRelease(imageRef);
        CGDataProviderRelease(dataProviderRef);
    }
}

- (void)setDim:(BOOL)value update:(BOOL)update
{
    fakeBoxRec bounds = {0, 0, [self frame].size.width,
                         [self frame].size.height};
                         
    if(mDimmed!=value) {
        mDimmed = value;
        if(update) [self refreshRects:&bounds count:1];
    }
}

- (BOOL)shadowChangeInhibited
{
    return mInhibitShadowChange;
}

- (void)setInhibitShadowChange:(BOOL)value
{
    mInhibitShadowChange = value;
}

- (BOOL)getMyMiniature
{
    return mMiniature;
}

- (void)setMyMiniature:(BOOL)value
{
    mMiniature = value;
}

- (void)setMyDrawAlpha:(int)alphaval
{
    mDrawAlpha = (unsigned long) alphaval << 24;
//    NSLog(@"Set alpha to %d",alphaval);
}

- (void)setMyTranslucentAlpha:(int)alphaval
{
    mTranslucentBG = mTranslucentBG & 0x00ffffff;
    mTranslucentBG = mTranslucentBG | ((unsigned long) alphaval << 24);
}

- (void)setMyTranslucentRed:(int)redval
{
    mTranslucentBG = mTranslucentBG & 0xffffff00;
    mTranslucentBG = mTranslucentBG | (redval & 0xff);
}

- (void)setMyTranslucentGreen:(int)greenval
{
    mTranslucentBG = mTranslucentBG & 0xffff00ff;
    mTranslucentBG = mTranslucentBG | (greenval << 8);
}

- (void)setMyTranslucentBlue:(int)blueval
{
    mTranslucentBG = mTranslucentBG & 0xff00ffff;
    mTranslucentBG = mTranslucentBG | ((unsigned long) blueval << 16);
}

- (void)setMyHackID:(int)hackid
{
    mHackID = hackid;
    mRectTag = [self addTrackingRect:[self frame] owner:self userData:nil assumeInside:NO];
    //NSLog(@"HACKID %d : created tracking rect... %d",mHackID,mRectTag);   
}

- (int)getMyHackID;
{
    return mHackID;
}

- (NSTrackingRectTag)getMyTrackingRect
{
    return mRectTag;
}

- (void)setMyTrackingRect:(NSTrackingRectTag)tag
{
    mRectTag=tag;
}

- (void)setInhibitFull:(BOOL)val
{
    mInhibitUpdate = val;
}

@end
